"""
Semantic Web Pidgin: Parse chart data structure
Max Goldman and Robin Stewart
"""

class Chart:
    """
    A "chart" to store possible parses of a sequence of tokens. A location
    (start, end) in the chart contains parses that explain tokens spanning
    [start,end] of the original input.
    """
    def __init__(self, tokens):
        """
        Create a new chart to store parses of the given tokens.
        """
        size = len(tokens)
        assert size > 0
        self._size = size
        self._parses = { }
        for i in range(size):
            self._parses[(i,i)] = [tokens[i]]

    def __len__(self):
        """
        Return the number of original tokens in this chart.
        """
        return self._size

    def possible_token_lists(self, start, end):
        """
        Generate a list of lists of tokens spanning [start,end].
        """
        token_lists = [ ]
        for token_bag_list in self._possible_token_bag_lists(start, end):
            token_lists += self._possible_token_lists(token_bag_list)
        return token_lists

    def _possible_token_lists(self, token_bag_list):
        """
        Given a list of bags of tokens, generate the list of lists of tokens
        that contains every way to pick one token from each bag.
        """
        token_lists = [ ]
        for token in token_bag_list[0]:
            if len(token_bag_list) > 1:
                for token_list in self._possible_token_lists(token_bag_list[1:]):
                    token_lists.append([ token ] + token_list)
            else:
                token_lists.append([ token ])
        return token_lists

    def _possible_token_bag_lists(self, start, end):
        """
        Generate the list of lists of bags of tokens spanning [start,end].
        """
        assert 0 <= start and start <= end and end < len(self)
        lists = [ [ self.parses(start, end) ] ]

        for split in range(start, end):
            right_parses = self.parses(split+1, end)
            if len(right_parses) > 0:
                for left_list in self._possible_token_bag_lists(start, split):
                    lists += [ left_list + [ right_parses ] ]
        return [ list for list in lists if [] not in list ]

    def parses(self, start, end):
        """
        Return the list of parses for the interval [start,end], or the empty
        list if there are none.
        """
        assert 0 <= start and start <= end and end < len(self)
        return self._parses.get((start,end), [])

    def add_parses(self, start, end, parses):
        """
        Add parses that explain the interval [start,end]. Returns true of any
        of the given parses is new, and false otherwise.
        """
        assert 0 <= start and start <= end and end < len(self)
        new_parses = False
        for parse in parses:
            if parse not in self.parses(start, end):
                ### print "new parse", start, end, parse
                self._parses[(start,end)] = self.parses(start, end) + [ parse ]
                new_parses = True
        return new_parses
