from __future__ import with_statement

import re
from rdflib import *
from rdflib.Graph import *
from rdflib.Namespace import *

"""
Semantic Web Pidgin: Parsing
Max Goldman and Robin Stewart
"""

#
# Basic classes
#

class Parser:
    def __init__(self):
        pass

    def parse(self, tokens):
        """
        Attempt to parse an array of Token objects and return an array of
        possible parses for these tokens.
        """
        pass

class Token:
    """
    A weighted, perhaps parsed, token of a particular type.
    """
    def __init__(self, type, object, weight):
        assert 0.0 <= weight <= 1.0
        self._type = type
        self._object = object
        self._weight = weight

    def __repr__(self):
        return "<Token=%s, type=%s, weight=%s>" % (self._object, self._type, self._weight)

    def type(self):
        return self._type

    def object(self):
        return self._object

    def weight(self):
        return self._weight

    def reweight(self, prior):
        assert 0.0 <= prior <= 1.0
        self._weight *= prior

#
# Semantic web starts here
#

RDF_TYPE = URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
RDF_RANGE = URIRef("http://www.w3.org/2000/01/rdf-schema#range")
RDF_STRING = URIRef("http://www.w3.org/2001/XMLSchema#string")
RDF_DATETIME = URIRef("http://www.w3.org/2001/XMLSchema#dateTime")
RDF_EMAIL = URIRef("http://people.csail.mit.edu/stewart/person.n3#Email")

NL_SURFACE = URIRef("http://people.csail.mit.edu/stewart/nl.n3#surface")
NL_PREP = URIRef("http://people.csail.mit.edu/stewart/nl.n3#prep")

class SemanticWebParser (Parser):
    """
    A parser that understands attempts to find existing or create new entities
    of a particular kind.
    """
    k_type = "type"
    k_prep = "preps"

    def __init__(self, store, model, entities=None):
        """
        Create a new parser configured according to the given model and
        optionally able to find entities in the given entities file. All RDF
        operations are performed in the context of the given datastore.
        """
        print "Semantic parser for", model
        self._store = store
        if entities:
            store.load(entities, format="n3")

        self._type = None
        self._properties = { }
        with open(model) as properties:
            for line in properties:
                property = line.strip()
                uriref = URIRef(property)
                if not self._type:
                    # the first line should define our type
                    store.load(property, format="n3")
                    self._type = uriref
                    self._surfaces = [ str(s) for s in store.objects(uriref, NL_SURFACE) ]
                else:
                    # remaining lines define our properties
                    self._properties[uriref] = { }
                    type = store.objects(uriref, RDF_RANGE).next()
                    self._properties[uriref][self.k_type] = type
                    preps = [ str(p) for p in store.objects(uriref, NL_PREP) ]
                    self._properties[uriref][self.k_prep] = preps
        assert self._type

        print "Can find using", entities
        for entity in self._store.subjects(RDF_TYPE, self._type):
            print " ", entity

        print "Can create using", self._type

    def parse(self, tokens):
        parses = [ ]
        tokens = self._attempt_consume_surface(tokens)
        if len(tokens) < 1:
            return parses
        parses += self._attempt_parse_as_find(tokens)
        parses += self._attempt_parse_as_create(tokens)
        return parses

    def _attempt_consume_surface(self, tokens):
        """
        Attempt to find a token with one of the possible surface forms and
        remove that token from the list.
        """
        for surface in self._surfaces:
            for token in tokens:
                if token.object() == surface:
                    return filter(lambda t: t != token, tokens)
        return tokens

    def _attempt_parse_as_find(self, tokens):
        """
        Attempt to parse the given tokens as referring to some entity.
        """
        parses = [ ]
        # for now, always parse by string matching
        if len(filter(lambda x: x.type() != RDF_STRING, tokens)) > 0:
            return parses
        for property in self._properties:
            parses += self._attempt_parse_as_find_property(tokens, property)
        return parses

    def _attempt_parse_as_find_property(self, tokens, property):
        """
        Attempt to parse the given tokens as referring to some entity by the
        given property. E.g., "karger" refers to the ...#name of a ...#Person.
        """
        # for now, assume all tokens contain strings
        parses = [ ]
        tokens = [token.object() for token in tokens]

        # check for preposition
        for prep in self._properties[property][self.k_prep]:
            if tokens[0] == prep:
                tokens = tokens[1:]
                break
        if len(tokens) < 1:
            return parses

        # match entities in the datastore
        tokenstring = reduce(lambda a,b: a+b, tokens)
        for subject in self._store.subjects(RDF_TYPE, self._type):
            for object in self._store.objects(subject, property):
                object = str(object)
                regex = reduce(lambda a,b: a+".*"+b, tokens)
                if re.search(regex, str(object), re.I):
                    weight = float(len(tokenstring))/len(str(object))
                    parses.append(Token(self._type, subject, weight))
        return parses

    def _attempt_parse_as_create(self, tokens):
        """
        Attempt to parse the given tokens as a request to create a new entity.
        """
        parses = [ ]

        # don't try to create single-property entities for performance reasons
        if len(self._properties) < 2:
            return parses

        object = { }
        object[RDF_TYPE] = self._type
        properties = [ p for p in self._properties ]
        objects = self._attempt_parse_as_create_recur([ object ], properties, tokens)

        weight = 1.0 - pow(0.5, len(tokens))
        for object in objects:
            parses.append(Token("CREATE A NEW " + self._type, object, weight))
        return parses

    def _attempt_parse_as_create_recur(self, prev_objects, properties, tokens):
        """
        Attempt to parse the given properties from the given tokens, extending
        the given objects to include the updated property value assignments.
        """
        if len(properties) < 1:
            if len(tokens) < 1:
                return prev_objects
            else:
                return [ ]

        objects = [ ]
        prop = properties[0]
        type = self._properties[prop][self.k_type]
        for token in filter(lambda t: t.type() == type, tokens):
            # consume one possible token or token sequence for the current
            # property
            p_left = properties[1:]
            t_left = filter(lambda t: t != token, tokens)
            prev_token = tokens[tokens.index(token)-1]
            for prep in self._properties[prop][self.k_prep]:
                if prev_token.object() == prep:
                    t_left.remove(prev_token)
                    break
            # update the objects
            updated_objects = [ o.copy() for o in prev_objects ]
            [ o.update({ prop: token.object() }) for o in updated_objects ]
            # recursively extend those objects according to the rest of the
            # phrase
            objects += self._attempt_parse_as_create_recur(updated_objects, p_left, t_left)
        return objects
